from collections import OrderedDict

from PySide import QtCore

from ..qt.QTreeSelectionModel import QTreeSelectionModel

class QOpManager(QTreeSelectionModel):
    """
    A QTreeSelectionModel for interacting with TreeModel OpManager.
    """

    def __init__(self,opman):
        default_flags = OrderedDict()
        default_flags['enable'] = False
        super(QOpManager,self).__init__(default_flags,opman)
        self.opman = opman

    def headerData(self,section,orientation,data_role):
        if (data_role == QtCore.Qt.DisplayRole and section == 0):
            return "{} operations available".format(self.opman.n_ops())
        else:
            return super(QOpManager,self).headerData(section,orientation,data_role) 

    def flags(self,idx):
        d = self.get_data_from_index(idx)
        if isinstance(d,dict) and not idx.column()==0:
            # Don't allow user to control enable flag on categories, at least for now
            # TODO: Consider allowing user to enable/disable entire categories
            #return QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsTristate
            return QtCore.Qt.NoItemFlags
        else:
            return super(QOpManager,self).flags(idx)

    # Override setData to handle operation enable/disable via check boxes. 
    def setData(self,idx,val,data_role):
        if idx.column() == 0:
            return super(QOpManager,self).setData(index,val,data_role)
        elif data_role == QtCore.Qt.CheckStateRole:
            itm = self.get_from_index(idx)
            op_uri = self.get_uri_of_index(idx)
            try:
                self.opman.set_op_enabled(op_uri,bool(val))
            except Exception as ex:
                msg = str('Failed to enable Operation {}. '.format(op_uri)
                + 'Error message: {}'.format(ex.message))
                self.opman.write_log(msg)
                return False
            self.set_flagged(itm,self.flag_defaults.keys()[idx.column()-1],val)
            self.dataChanged.emit(idx,idx)
            return True
        else:
            return super(QTreeSelectionModel,self).setData(index,val,data_role)


    #def idx_of_cat(self,catname,p_idx):
    #    """
    #    If cat exists under p_idx, return its index, 
    #    else return an invalid QModelIndex.
    #    """
    #    ncats = self.item_count(p_idx)
    #    p_itm = self.get_from_idx(p_idx)
    #    for j in range(ncats):
    #        idx = self.index(j,0,p_idx)
    #        cat = self.get_from_idx(idx).tag
    #        if cat == catname:
    #            return idx
    #    return QtCore.QModelIndex() 

    #def remove_ops(self,rm_idx):
    #    """
    #    Remove an Operation or category from the tree.
    #    """
    #    if not rm_idx == self.root_index() and rm_idx.isValid():
    #        p_idx = self.parent(rm_idx)
    #        rm_itm = self.get_from_idx(rm_idx)
    #        self.delete_item(rm_itm.tag,p_idx)            

    # get an Operation by its cat.[subcat.(...).]opname uri
    #def get_op_from_uri(self,op_uri):
        #return self.get_data_from_uri(op_uri)
        #for op in self._op_list:
        #    if op.__name__ == op_name:
        #        return op
        #return None

    # get an Operation from the list by its TreeItem's QModelIndex
    #def get_op(self,indx):
        #treeitem = self.get_item(indx)
        #return treeitem.data
 





"""
paws setup module
---- ----- ------
from https://packaging.python.org/distributing/
"""

#from setuptools import setup, find_packages
# To use a consistent encoding
#from codecs import open

from os import path
from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here,'README.rst')) as f:
    long_description = f.read()

# Executing paws_config.py defines __version__ 
with open(path.join(here,'paws/paws_config.py')) as f: 
    exec(f.read())

__authorship__=''
here = path.abspath(path.dirname(__file__))

with open(path.join(here,'contributors.txt')) as f: 
    for line in f.readlines():
        __authorship__ += line.strip()+', '
__authorship__ = __authorship__[:-2]


setup(
    name='pypaws',

    # Versions should comply with PEP440.  
    version=__version__,

    description='the Platform for Automated Workflows by SSRL',
    long_description=long_description,

    url='https://github.com/slaclab/paws/',
    author=__authorship__,
    author_email='paws-developers@slac.stanford.edu',
    license='BSD',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[

        'Development Status :: 3 - Alpha',
        #'Development Status :: 4 - Beta',
        #'Development Status :: 5 - Production/Stable',
        #'Development Status :: 6 - Mature',

        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: BSD License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2.7',
    ],

    keywords='data analysis workflow',

    packages=find_packages(exclude=[]),
    #py_modules=["my_module"],

    package_dir={'paws':'paws'},

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['pyyaml'],
    python_requires='>=2.6, <3',

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    #extras_require={
    #    'dev': ['check-manifest'],
    #    'test': ['coverage'],
    #},

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    package_data={
        'paws': ['ui/graphics/*.png','ui/qtui/*.ui'],
    },

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    data_files=[('', ['contributors.txt'])],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    # TODO: entry point for paws python console
    entry_points={
        'console_scripts': [
            'paws=paws:main',
        ],
    },
)



"""
Created on Mon Jun 06 18:02:32 2016

author(s): Fang Ren, Apurva Mehta
Module originally contributed by Fang Ren.
For details, refer to the recent paper submitted to ACS Combinatorial Science.
TODO: Get this citation
"""

import numpy as np

from ... import Operation as op
from ...Operation import Operation

class TextureFeatures(Operation):
    """
    Analyze the texture 
    """

    def __init__(self):
        input_names = ['q','chi','I']
        output_names = ['q_texture','texture','int_sqr_texture']
        super(TextureFeatures,self).__init__(input_names,output_names)
        self.input_doc['q'] = '1d array of momentum transfer values'
        self.input_doc['chi'] = '1d array of out-of-plane diffraction angles'
        self.input_doc['I'] = '2d array representing intensities at q,chi points'
        self.input_src['q'] = op.wf_input
        self.input_src['chi'] = op.wf_input
        self.input_src['I'] = op.wf_input 
        self.input_type['q'] = op.ref_type
        self.input_type['chi'] = op.ref_type
        self.input_type['I'] = op.ref_type 
        self.output_doc['q_texture'] = 'q values at which the texture is analyzed'
        self.output_doc['texture'] = 'quantification of texture for each q'
        self.output_doc['int_sqr_texture'] = 'integral over q of the texture squared'

    def run(self):
        q, chi = np.meshgrid(self.inputs['q'], self.inputs['chi']*np.pi/float(180))
        keep = (self.inputs['I'] != 0)
        I = keep.astype(int) * self.inputs['I']
        # TODO: This appears to be a binning operation.
        # Maybe the bin size should not be hard-coded. 
        I_sum = np.bincount((q.ravel()*100).astype(int), I.ravel().astype(int))
        count = np.bincount((q.ravel()*100).astype(int), keep.ravel().astype(int))
        I_ave = np.array(I_sum)/np.array(count)
        texsum = np.bincount((q.ravel()*100).astype(int), (I*np.cos(chi)).ravel())
        chi_count = np.bincount((q.ravel()*100).astype(int), (keep*np.cos(chi)).ravel())
        texture = np.array(texsum) / np.array(I_ave) / np.array(chi_count) - 1
        step = 0.01
        q_texture = np.arange(step,np.max(q)+step,step)
        tsqr_int = np.nansum(texture ** 2)/float(q_texture[-1]-q_texture[0])
        self.outputs['q_texture'] = q_texture
        self.outputs['texture'] = texture
        self.outputs['int_sqr_texture'] = tsqr_int 






import numpy as np

from .. import Operation as op
from ..Operation import Operation

class BigLoad(Operation):
    """An Operation testing class, creates and outputs a big array of noise"""

    def __init__(self):
        input_names = ['size']
        output_names = ['big_array']
        super(BigLoad,self).__init__(input_names,output_names) 
        self.input_doc['size'] = 'dimension of big array'
        self.output_doc['big_array'] = 'a big array of noise'
        self.input_src['size'] = op.text_input    
        self.input_type['size'] = op.int_type
        self.inputs['size'] = 10000 
    
    def run(self):
        s = self.inputs['size']
        self.outputs['big_array'] = np.random.rand(s,s) 




"""
Operation for automatic discovery of diffraction image calibration parameters. 
"""

import numpy as np
import pyFAI as pf
import pyFAI.calibrant as pfc

from ... import Operation as op
from ...Operation import Operation

class AutoCal(Operation):
    """
    Input image data (ndarray), pixel size, polz factor.
    Output dictionary of calibration parameters. 
    """
    def __init__(self):
        input_names = ['image_data','pixel_size','fpolz']
        output_names = ['cal_params']
        super(AutoCal,self).__init__(input_names,output_names)
        self.input_doc['image_data'] = '2d array representing intensity for each pixel'
        self.input_doc['calibrant'] = 'name of calibrant- supported values: AgBh'
        self.input_doc['wavelength'] = 'wavelength, in Angstrom'
        self.input_doc['pixel_size'] = 'pixel size in microns'
        self.input_doc['fpolz'] = 'polarization factor'
        self.input_src['image_data'] = op.wf_input
        self.input_src['calibrant'] = op.text_input 
        self.input_src['wavelength'] = op.text_input 
        self.input_src['pixel_size'] = op.text_input 
        self.input_src['fpolz'] = op.text_input
        self.input_type['image_data'] = op.ref_type
        self.input_type['calibrant'] = op.str_type
        self.input_type['wavelength'] = op.float_type
        self.input_type['pixel_size'] = op.float_type
        self.input_type['fpolz'] = op.float_type
        self.inputs['calibrant'] = 'AgBh' 
        self.inputs['wavelength'] = 0.799898 
        self.inputs['pixel_size'] = 79 
        self.inputs['fpolz'] = 0.95 
        self.output_doc['cal_params'] = 'Dictionary of the solved calibration parameters, .poni format'

    def run(self):
        img = self.inputs['image_data']
        calid = self.inputs['calibrant']
        wl = self.inputs['wavelength'] 
        pxsz = self.inputs['pixel_size']
        fpolz = self.inputs['fpolz']
        if calid == 'AgBh':
            # Define an AgBh calibrant
            cal = pfc.ALL_CALIBRANTS[calid]
            cal.set_wavelength(1E-10)
        else:
            raise ValueError('[{}] Unsupported calibrant: {}'.format(__name__,calid))

        x0_px = 0
        y0_px = 0
        d_px = 0
        rot_rad = 0
        tilt_rad = 0
        wl = 0
        d = dict({'x0_pixel':x0_px,
            'y0_pixel':y0_px,
            'd_pixel':d_px,
            'pixel_size':pxsz,
            'rotation_rad':rot_rad,
            'tilt_rad':tilt_rad,
            'lambda':wl,
            'PP':fpolz})
        self.outputs['cal_params'] = d





import importlib
from functools import partial

from PySide import QtCore, QtGui, QtUiTools

from ..core.models.ListModel import PluginListModel
from ..core import plugins 
from ..core import pawstools
from ..core.plugins.PawsPlugin import PawsPlugin
from ..core.operations import Operation as op
from ..core.operations import optools
from .InputLoader import InputLoader
from . import uitools

class PluginUiManager(QtCore.QObject):
    # TODO: API cleanup: hide refs to self.qplugman.plugman._tree

    def __init__(self,qplugman):
        ui_file = QtCore.QFile(pawstools.sourcedir+"/ui/qtui/plugin_control.ui")
        ui_file.open(QtCore.QFile.ReadOnly)
        self.ui = QtUiTools.QUiLoader().load(ui_file)
        ui_file.close()
        # set self.ui to be deleted and to emit destroyed() signal when its window is closed
        self.ui.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        self.qplugman = qplugman 
        self.plugin_list_model = PluginListModel(plugins.plugin_name_list,self.ui)
        self.pgin = None
        self.setup_ui()
        self.src_widgets = {} 
        self.type_widgets = {} 
        self.val_widgets = {} 
        self.btn_widgets = {} 
        self.input_loaders = {} 
        self.name_col = 0
        self.eq_col = 1
        self.src_col = 2
        self.type_col = 3
        self.val_col = 4
        self.btn_col = 5
        self.invalid_sources = [op.batch_input,op.plugin_input,op.wf_input]

    def setup_ui(self):
        self.ui.setWindowTitle("plugin setup")
        self.ui.input_box.setTitle("SETUP")
        self.ui.finish_box.setTitle("FINISH / LOAD")
        ht = self.ui.plugin_frame.sizeHint().height()
        self.ui.plugin_frame.sizeHint = lambda: QtCore.QSize(300,ht)
        self.ui.plugin_frame.setSizePolicy(
        QtGui.QSizePolicy.Minimum,self.ui.plugin_frame.sizePolicy().verticalPolicy())
        self.ui.plugins_available.clicked.connect( partial(self.start_plugin) )
        self.ui.plugins_available.setModel(self.plugin_list_model)
        self.ui.plugins_loaded.setModel(self.qplugman)
        self.ui.plugins_loaded.hideColumn(1)
        self.ui.plugins_loaded.hideColumn(2)
        self.ui.plugins_loaded.setRootIndex(self.qplugman.root_index())
        self.ui.stop_plugin_button.clicked.connect(self.stop_plugin)
        self.ui.stop_plugin_button.setText("&Stop selected plugin")
        self.ui.uri_prompt.setMaximumWidth(150)
        self.ui.uri_prompt.setText('plugin tag:')
        self.ui.plugin_name.setAlignment(QtCore.Qt.AlignCenter)
        self.ui.plugin_name.setText('-select a plugin to begin setup-')
        self.ui.uri_prompt.setAlignment(QtCore.Qt.AlignRight)
        self.ui.uri_prompt.setStyleSheet( "QLineEdit { background-color: transparent }" 
        + self.ui.uri_prompt.styleSheet() )
        self.ui.load_button.setText("&Finish")
        self.ui.load_button.setMinimumWidth(100)
        self.ui.load_button.clicked.connect(self.load_plugin)
        self.ui.load_button.setEnabled(False)
        #self.ui.splitter.setStretchFactor(0,1000)    
        self.ui.setStyleSheet( "QLineEdit { border: none }" + self.ui.styleSheet() )

    def start_plugin(self,idx):
        pkg = plugins.__name__
        pgin_name = self.plugin_list_model.get_item(idx)
        #print 'start plugin {} from package {}'.format(pgin_name,pkg)
        mod = importlib.import_module('.'+pgin_name,pkg)
        for nm, itm in mod.__dict__.items():
            if isinstance(itm,type):
                if issubclass(itm,PawsPlugin) and not nm == 'PawsPlugin':
                    pgin = getattr(mod,nm)
                    self.set_plugin(pgin())
                    return

    def set_plugin(self,pgin):
        self.pgin = pgin
        self.ui.plugin_info.setPlainText(self.pgin.description())
        self.build_input()
        # TODO: clean up the following
        self.ui.uri_entry.setText(
        self.qplugman.plugman._tree.make_unique_uri(type(self.pgin).__name__))
        
    def clear_input(self):
        self.ui.plugin_name.setText('')
        self.ui.uri_entry.setText('')
        self.ui.load_button.setEnabled(False)
        n_inp_widgets = self.ui.input_layout.count()
        for i in range(n_inp_widgets-1,-1,-1):
            item = self.ui.input_layout.takeAt(i)
            item.widget().close()
        self.src_widgets = {}
        self.type_widgets = {}
        self.val_widgets = {}
        self.btn_widgets = {}
        self.input_loaders = {}

    def reset_input_headers(self):
        if len(self.pgin.inputs) > 0:
            self.ui.input_layout.addWidget(uitools.r_hdr_widget('name'),0,self.name_col,1,1)
            self.ui.input_layout.addWidget(uitools.hdr_widget('source'),0,self.src_col,1,1)
            self.ui.input_layout.addWidget(uitools.hdr_widget('type'),0,self.type_col,1,1)
            self.ui.input_layout.addWidget(uitools.hdr_widget('value'),0,self.val_col,1,1)

    def build_input(self):
        self.clear_input()
        self.reset_input_headers()
        self.ui.plugin_name.setText(type(self.pgin).__name__)
        self.ui.load_button.setEnabled(True)
        for i,name in zip(range(1,len(self.pgin.inputs)+1),self.pgin.inputs.keys()):
            # name 
            name_widget = uitools.name_widget(name)
            self.ui.input_layout.addWidget( name_widget,i,self.name_col,1,1 )
            # '=' 
            eq_widget = uitools.smalltext_widget('=')
            self.ui.input_layout.addWidget(eq_widget,i,self.eq_col,1,1)
            # source 
            src_widget = uitools.src_selection_widget()
            for s in self.invalid_sources:
                src_widget.model().set_disabled(s)
            src = self.pgin.input_src[name]
            if src in self.invalid_sources:
                src = op.no_input
            src_widget.setCurrentIndex(src)
            self.ui.input_layout.addWidget( src_widget,i,self.src_col,1,1 )
            self.src_widgets[name] = src_widget 
            self.src_widgets[name].currentIndexChanged.connect( partial(self.reset_type_widget,name,i) )
            # type 
            self.reset_type_widget(name,i,src)
            #tp = self.type_widgets[name].currentIndex()
            # val 
            #self.reset_val_widget(name,i,src,tp)

    def reset_type_widget(self,name,row,src=None):
        if name in self.type_widgets.keys():
            if self.type_widgets[name]:
                self.type_widgets[name].close()
        if not src:
            src = self.src_widgets[name].currentIndex()
        type_widget = uitools.type_selection_widget(src)
        if self.pgin.input_type[name]:
            if self.pgin.input_type[name] not in op.invalid_types[src]:
                type_widget.setCurrentIndex(self.pgin.input_type[name])
            else:
                # set sensible defaults
                if src == op.fs_input:
                    new_type_widget.setCurrentIndex(op.path_type)
                elif src == op.text_input:
                    new_type_widget.setCurrentIndex(op.str_type)
        #if type_widget.currentIndex() in op.invalid_types[src]:
        #    type_widget.setCurrentIndex(op.none_type)
        #if src in [op.wf_input,op.plugin_input]:
        #    type_widget.setCurrentIndex(op.ref_type)
        #if src == op.fs_input:
        #    type_widget.setCurrentIndex(op.path_type)
        type_widget.currentIndexChanged.connect( partial(self.reset_val_widget,name,row,src) )            
        self.type_widgets[name] = type_widget  
        self.ui.input_layout.addWidget(type_widget,row,self.type_col,1,1)
        tp = self.type_widgets[name].currentIndex()
        self.reset_val_widget(name,row,src,tp)

    def reset_val_widget(self,name,row,src=None,tp=None):
        # TODO: Make val widgets expand height according to their content 
        if name in self.val_widgets.keys():
            if self.val_widgets[name]:
                self.val_widgets[name].close()
        if name in self.btn_widgets.keys():
            if self.btn_widgets[name]:
                self.btn_widgets[name].close()
        if name in self.input_loaders.keys():
            if self.input_loaders[name]:
                self.input_loaders[name].ui.close()
                self.input_loaders[name] = None
        if not src:
            src = self.src_widgets[name].currentIndex()
        if not tp:
            tp = self.type_widgets[name].currentIndex()
        btn_widget = QtGui.QPushButton()
        val_widget = QtGui.QLineEdit()
        if src == op.no_input or tp == op.none_type: 
            btn_widget.setText('no input')
            btn_widget.setEnabled(False)
            val_widget.setText('None')
        #elif src in [op.wf_input,op.fs_input,op.plugin_input,op.text_input]:
        elif src in [op.fs_input,op.text_input]:
            btn_widget.setText('edit...')
            btn_widget.clicked.connect( partial(self.fetch_data,name) )
            if self.pgin.inputs[name] is not None:
                val_widget.setText(str(self.pgin.inputs[name]))
        val_widget.setReadOnly(True)
        self.ui.input_layout.addWidget(val_widget,row,self.val_col,1,1)
        self.ui.input_layout.addWidget(btn_widget,row,self.btn_col,1,1)
        self.val_widgets[name] = val_widget
        self.btn_widgets[name] = btn_widget
       
    def fetch_data(self,name):
        if name in self.input_loaders.keys():
            if self.input_loaders[name]:
                self.input_loaders[name].ui.close()
                self.input_loaders[name] = None
        src = self.src_widgets[name].currentIndex()
        if src == op.fs_input:
            trmod = QtGui.QFileSystemModel()
            inp_loader = InputLoader(name,src,trmod,self.ui)
        elif src == op.text_input:
            inp_loader = InputLoader(name,src,None,self.ui)
        if self.pgin.input_src[name] == src and self.pgin.inputs[name] is not None:
            if isinstance(self.pgin.inputs[name],list):
                inp_loader.set_list_toggle()
                for v in self.pgin.inputs[name]:
                    inp_loader.add_value(str(v))
            else:
                inp_loader.add_value(str(self.pgin.inputs[name]))
        inp_loader.ui.finish_button.clicked.connect( partial(self.set_input,name,inp_loader.ui) )
        self.input_loaders[name] = inp_loader
        inp_loader.ui.show()

    def fetch_from_input_ui(self,ui):
        # no matter the input source, ui.values_list.model().list_data() should be a list of strings.
        val = ui.values_list.model().list_data()
        # if ui.list_toggle is not checked, the value should be unpacked. 
        if not ui.list_toggle.isChecked():
            if len(val) == 0:
                val = None
            else:
                val = val[0]
        return val

    def set_input(self,name,src_ui=None):
        src = self.src_widgets[name].currentIndex()
        tp = self.type_widgets[name].currentIndex()
        if src_ui:
            val = self.fetch_from_input_ui(src_ui) 
            self.pgin.inputs[name] = val 
            self.val_widgets[name].setText(val)
            src_ui.close()
            # dereference the ui now that it is closed...
            self.input_loaders[name] = None
            #src_ui.deleteLater()
        elif src == op.text_input:
            val = self.val_widgets[name].text()
            self.pgin.inputs[name] = optools.cast_type_val(tp,val)

    def load_plugin(self):
        """
        Package the finished Plugin, ship to plugin manager
        """ 
        for name in self.pgin.inputs.keys():
            self.set_input(name)
        uri = self.ui.uri_entry.text()
        # Plugin setup occurs here via PawsPlugin.start()
        self.pgin.start()
        if self.qplugman.plugman.is_uri_valid(uri):
            self.qplugman.add_plugin(uri,self.pgin) 
            self.clear_input()
        else:
            # Request a different uri 
            msg_ui = uitools.message_ui(self.ui)
            msg_ui.setWindowTitle("Tag Error")
            msg_ui.message_box.setPlainText(
            self.qplugman.plugman._tree.tag_error(uri))
            msg_ui.setAttribute(QtCore.Qt.WA_DeleteOnClose)
            msg_ui.show()

    def stop_plugin(self,idx=QtCore.QModelIndex()):
        """
        remove the selected plugin from self.plugman 
        """
        if not idx.isValid():
            idx = self.ui.plugins_loaded.currentIndex()
        if idx.isValid(): 
            while idx.internalPointer().parent.isValid():
                idx = idx.internalPointer().parent
            pgin_tag = self.qplugman.get_uri_of_index(idx)
            self.qplugman.remove_plugin(pgin_tag)





import numpy as np

from ... import Operation as op
from ...Operation import Operation
from ... import optools

class XYDataFromBatch(Operation):
    """
    Given a batch output and appropriate keys, 
    use the uris to harvest x and y data from the batch.
    """

    def __init__(self):
        input_names = ['batch_output','x_uri','y_uri','x_shift_flag']
        output_names = ['x','y','x_y','x_y_sorted']
        super(XYDataFromBatch,self).__init__(input_names,output_names)        
        self.input_doc['batch_output'] = 'list of dicts produced by a batch execution.'
        self.input_doc['x_uri'] = 'uri of data for x. Must be in batch.saved_items().'
        self.input_doc['y_uri'] = 'uri of data for y. Must be in batch.saved_items().'
        self.input_doc['x_shift_flag'] = 'if True, shift x data so that its minimum value is zero.' 
        self.input_src['batch_output'] = op.wf_input
        self.input_src['x_uri'] = op.wf_input
        self.input_src['y_uri'] = op.wf_input
        self.input_src['x_shift_flag'] = op.text_input
        self.input_type['batch_output'] = op.ref_type
        self.input_type['x_uri'] = op.path_type
        self.input_type['y_uri'] = op.path_type
        self.input_type['x_shift_flag'] = op.bool_type
        self.inputs['x_shift_flag'] = False
        self.output_doc['x'] = 'array of the x values in batch output order.'
        self.output_doc['y'] = 'array of the y values in batch output order.'
        self.output_doc['x_y'] = 'n-by-2 array of x and y values in batch output order.'
        self.output_doc['x_y_sorted'] = 'n-by-2 array of x and y values, sorted for increasing x.'

    def run(self):
        b_out = self.inputs['batch_output']
        x_uri = self.inputs['x_uri']
        y_uri = self.inputs['y_uri']
        x_all = np.array([optools.get_uri_from_dict(x_uri,d) for d in b_out],dtype=float)
        y_all = np.array([optools.get_uri_from_dict(y_uri,d) for d in b_out],dtype=float)
        if any(x_all):
            xmin = np.min(x_all)
        else:
            xmin = 0
        if self.inputs['x_shift_flag']:
            x_all = x_all - xmin
            #xmin = 0 
        self.outputs['x'] = x_all 
        self.outputs['y'] = y_all 
        self.outputs['x_y'] = np.array(zip(x_all,y_all))
        #self.outputs['x_y_sorted'] = np.sort(np.array(zip(x_all,y_all)),0)
        x_sort = np.sort(x_all)
        y_xsort = y_all[np.argsort(x_all)]
        self.outputs['x_y_sorted'] = np.array(zip(x_sort,y_xsort))





import numpy as np
from PIL import Image

from ... import Operation as op
from ...Operation import Operation

class LoadTif_PIL(Operation):
    """
    Takes a filesystem path that points to a .tif,
    outputs image data and metadata from the file. 
    """

    def __init__(self):
        input_names = ['path']
        output_names = ['image_data','metadata']
        super(LoadTif_PIL,self).__init__(input_names,output_names)
        self.input_doc['path'] = 'path to a .tif image'
        self.input_src['path'] = op.fs_input
        self.input_type['path'] = op.path_type
        self.output_doc['image_data'] = '2D array representing pixel values'
        self.output_doc['metadata'] = 'Dictionary of image metadata'
        
    def run(self):
        img_url = self.inputs['path']
        try:
            pil_img = Image.open(img_url)
            self.outputs['image_data'] = np.array(pil_img)
            self.outputs['metadata'] = pil_img.info
        except IOError as ex:
            ex.message = "[{}] IOError for file {}. \nError message:".format(__name__,img_url,ex.message)
            raise ex
        except ValueError as ex:
            ex.message = "[{}] ValueError for file {}. \nError message:".format(__name__,img_url,ex.message)
            raise ex

